import os
import subprocess
import time
from pathlib import Path
from random import choice
from string import ascii_lowercase

from run_iocsh import IOC

# Random module name
TEST_MODULE_NAME = "test_mod_" + "".join(choice(ascii_lowercase) for _ in range(16))

EPICS_BASE = os.environ.get("EPICS_BASE")
E3_REQUIRE_VERSION = os.environ.get("E3_REQUIRE_VERSION")


def set_env_vars(env: dict) -> dict:
    """
    Set a number of environment variables needed for require to build a module
    """
    assert "EPICS_BASE" in env
    assert Path(env["EPICS_BASE"]).is_dir
    assert "E3_REQUIRE_VERSION" in env
    assert env["E3_REQUIRE_VERSION"]

    E3_REQUIRE_LOCATION = f"{EPICS_BASE}/require/{E3_REQUIRE_VERSION}"
    assert Path(E3_REQUIRE_LOCATION).is_dir()
    env["E3_REQUIRE_LOCATION"] = E3_REQUIRE_LOCATION

    REQUIRE_CONFIG = f"{E3_REQUIRE_LOCATION}/configure"
    env["REQUIRE_CONFIG"] = REQUIRE_CONFIG

    env["E3_MODULE_NAME"] = TEST_MODULE_NAME
    env["E3_MODULE_SRC_PATH"] = TEST_MODULE_NAME
    env["E3_MODULE_MAKEFILE"] = f"{TEST_MODULE_NAME}.Makefile"

    # Add a default version
    env["E3_MODULE_VERSION"] = "0.0.0"

    return env


def run_make(path, *args, **kwargs):
    """
    Attempt to run `make <args>` in the specified path with <kwargs> as environment variables
    """
    test_env = set_env_vars(os.environ.copy())
    for kw in kwargs:
        test_env[kw] = kwargs[kw]
    make_cmd = ["make", "-C", path] + list(args)
    p = subprocess.Popen(
        make_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=test_env,
        encoding="utf-8",
    )
    outs, errs = p.communicate()
    return p.returncode, outs, errs


def get_env_var(path, var: str):
    """
    Fetch an environment variable from the module build environment
    """
    _, out, _ = run_make(path, "cellvars")
    for line in out.split("\n"):
        if line.startswith(var):
            return line.split("=")[1].strip()
    return ""


def run_ioc_get_output(version, cell_path):
    """
    Run an IOC and try to load the test module
    """
    with IOC("-r", f"{TEST_MODULE_NAME},{version}", "-l", cell_path) as ioc:
        time.sleep(1)
    return ioc.proc.returncode, ioc.outs, ioc.errs
